static char ident[] = "@(#)$Header: x11fish.c 1.1 91/10/28 $";


#include <sys/types.h>
#include <sys/time.h>
#include <stdio.h>
#include <signal.h>
#include <X11/Xlib.h>
#include "x11fish.h"

#ifdef siemens
# define sigvector sigvec
#endif

/* constants based on rand(3C) returning an integer between 0 and 32767 */
#define  RAND_F_1_8    4096.0
#define  RAND_I_1_4    8192
#define  RAND_I_1_2   16384
#define  RAND_I_3_4   24575
#define  RAND_F_MAX   32767.0


/* externals for pixmap and bimaps from xfishy.c */
extern rasinfo  xinfo;
extern char     xfishRaster[], xbBits[9][8];
extern int      binc[];


char           *pname,         /* program name from argv[0]                  */
                cname[64];     /* colorname specification                    */
int             sid,           /* default screen number                      */
                width,         /* width of initial window in pixels          */
                height,        /* height of initial window in pixels         */
                blimit = 24,   /* bubble limit                               */
                flimit = 4,    /* fish limit                                 */
                pmode = 1;     /* pop mode, (1 for lower, 0 for raise)       */
unsigned long  *cmap;          /* colormap vector, allocated dynamicaly      */
double          delay = 0.4,   /* update interval in seconds                 */
                smooth = 1.0;  /* smoothness increment multiplier            */
rasinfo         rinfo[2];      /* fish raster info structures                */
bubble         *binfo;         /* bubble info structs, allocated dynamically */
fish           *finfo;         /* fish info structs, allocated dynamically   */
GC              fgc,           /* fish graphic context                       */
                bgc1, bgc2;    /* bubble graphic contexts                    */
Pixmap          xfish[5],      /* fish pixmaps (element 0 is unused)         */
                xbubbles[9];   /* bubbles bitmaps (1 to 8 by size in pixels) */
Display        *did;           /* display info structure                     */
Window          wid;           /* aqaurium window                            */


#ifdef siemens
# define myrand() (rand() & 0x7FFF)
#else
# define myrand() rand()
#endif


/* ===========================================================================
Set up program defaults, get X defaults, parse command line using getopts.
Curiously enough, also open the display (required for XGetDefault's).
=========================================================================== */
void parse(argc, argv)
int     argc;              
char  **argv;
{
   int            c;
   char          *p, dname[64];
   extern int     optind;
   extern char   *optarg;
   extern double  atof();

   strcpy(dname, getenv("DISPLAY"));
   if (argc > 1) if (argv[argc-1][0] != '-') strcpy(dname, argv[argc-1]);
   if ((did = XOpenDisplay(dname)) == NULL) {
      fprintf(stderr, "%s: XOpenDisplay failed for display '%s'\n", pname, dname);
      exit(1);
   }

   strcpy(cname, "MediumSeaGreen");

   if ((p = XGetDefault(did, pname, "Bubbles")) != NULL) blimit = atoi(p);
   if ((p = XGetDefault(did, pname, "Color")) != NULL)   strcpy(cname, p);
   if ((p = XGetDefault(did, pname, "Delay")) != NULL)   delay  = atof(p);
   if ((p = XGetDefault(did, pname, "Fishes")) != NULL)  flimit = atoi(p);
   if ((p = XGetDefault(did, pname, "Smooth")) != NULL)  smooth = atof(p);

   while ((c = getopt(argc, argv, "b:c:d:f:s:x")) != EOF) {
      switch(c) {
      case 'b' :
         blimit = atoi(optarg);  break;
      case 'c' :
         strcpy(cname, optarg);  break;
      case 'd' :
         delay  = atof(optarg);  break;
      case 'f' :
         flimit = atoi(optarg);  break;
      case 's' :
         smooth = atof(optarg);  break;
      case 'x' :
         pmode  = 0;             break;
      case '?' :
         fprintf(stderr, "usage: %s [-b num][-c color][-d sec][-f num][-s mult][-x][display_name]\n", pname);
         exit(1);
      }
   }
}


/* ===========================================================================
Initialize colormap for background color and required fish colors.
The fish colors are coded in xfishy.c as a trio of tables.
=========================================================================== */
void init_colormap()
{
   int     i;
   XColor  sdef, edef;

   cmap = (unsigned long *)malloc((xinfo.colors + 1) * sizeof(unsigned long));

   XAllocNamedColor(did, XDefaultColormap(did, sid), cname, &sdef, &edef);
   cmap[0] = sdef.pixel;

   for (i = 0; i < xinfo.colors; i++) {
      sdef.red   = xinfo.reds[i];
      sdef.green = xinfo.greens[i];
      sdef.blue  = xinfo.blues[i];
      XAllocColor(did, XDefaultColormap(did, sid), &sdef);
      cmap[i+1]  = sdef.pixel;
   }
   cmap[xinfo.back_color] = cmap[0];
}


/* ===========================================================================
Calibrate the bubble pixmaps.  The eight bubbles are coded in xfishy.c as a
two dimensional array of char.
=========================================================================== */
void init_bubbles()
{
   int  i, depth;

   bgc1 = XCreateGC(did, wid, 0x0L, NULL);
   XSetGraphicsExposures(did, bgc1, False);
   XSetState(did, bgc1, 0L, ~0L, GXand, AllPlanes);

   bgc2 = XCreateGC(did, wid, 0x0L, NULL);
   XSetGraphicsExposures(did, bgc2, False);
   XSetState(did, bgc2, WhitePixel(did, sid), 0L, GXor, AllPlanes);
   
   depth = XDefaultDepth(did, sid);
   for (i = 1; i <= 8; i++)
      xbubbles[i] = XCreatePixmapFromBitmapData(did, wid, xbBits[i], i, i, 1L, 0L, 1);

   binfo = (bubble *)malloc(blimit * sizeof(bubble));
}


/* ===========================================================================
Calibrate the fish pixmaps.  The left-fish data is coded in xfishy.c, this
is transformed to create the right-fish.  These are then scaled up to create
the larger fish.
=========================================================================== */
void init_fish()
{
   register char  *p, *q, *x1, *x2;
   register int    i, j;
   int             depth;
   XImage          image;

   fgc = XCreateGC(did, wid, 0x0L, NULL);
   XSetGraphicsExposures(did, fgc, False);

   rinfo[0] = rinfo[1] = xinfo;
   rinfo[1].width *= 2;
   rinfo[1].height *= 2;

   x1 = xfishRaster;
   x2 = (char *)malloc(rinfo[1].width * rinfo[1].height * sizeof(char));

   depth = XDefaultDepth(did, sid);
   for (i = 1; i <= 4; i++) {
      if (i < 3) j = 0; else j = 1;
      xfish[i] = XCreatePixmap(did, wid, rinfo[j].width, rinfo[j].height, depth);
   }

   image.width            = rinfo[0].width;
   image.height           = rinfo[0].height;
   image.format           = ZPixmap;
   image.byte_order       = XImageByteOrder(did);
   image.bitmap_unit      = XBitmapUnit(did);
   image.bitmap_bit_order = XBitmapBitOrder(did);
   image.bitmap_pad       = XBitmapPad(did);
   image.depth            = 8 * sizeof(char);
   image.bytes_per_line   = rinfo[0].width;
   image.bits_per_pixel   = 8 * sizeof(char);

   j = rinfo[0].width * rinfo[0].height;
      for (i = 0, p = x1; i < j; i++, p++) *p = cmap[*p];

   image.data = x1;
   XPutImage(did, xfish[1], fgc, &image, 0, 0, 0, 0, rinfo[0].width, rinfo[0].height); 

   for (i = 0; i < rinfo[0].height; i++) {
      p = x1 + i * rinfo[0].width;
      q = x2 + (i + 1) * rinfo[0].width - 1;
      for (j = 0; j < rinfo[0].width; j++) *q-- = *p++;
   }
   image.data = x2;
   XPutImage(did, xfish[2], fgc, &image, 0, 0, 0, 0, rinfo[0].width, rinfo[0].height);

   for (i = 0; i < rinfo[0].height; i++) {
      p = x1 +  i * rinfo[0].width;
      q = x2 + i * 2 * rinfo[1].width;
      for (j = 0; j < rinfo[0].width; j++) { *q++ = *p;  *q++ = *p++; }
      p = x1 +  i * rinfo[0].width;
      q = x2 + (i * 2 + 1) * rinfo[1].width;
      for (j = 0; j < rinfo[0].width; j++) { *q++ = *p;  *q++ = *p++; }
   }
   image.width          = rinfo[1].width;
   image.height         = rinfo[1].height;
   image.bytes_per_line = rinfo[1].width;
   XPutImage(did, xfish[3], fgc, &image, 0, 0, 0, 0, rinfo[1].width, rinfo[1].height);

   for (i = 0; i < rinfo[0].height; i++) {
      p = x1 +  i * rinfo[0].width;
      q = x2 + (i * 2 + 1) * rinfo[1].width - 1;
      for (j = 0; j < rinfo[0].width; j++) { *q-- = *p;  *q-- = *p++; }
      p = x1 +  i * rinfo[0].width;
      q = x2 + (i * 2 + 2) * rinfo[1].width - 1;
      for (j = 0; j < rinfo[0].width; j++) { *q-- = *p;  *q-- = *p++; }
   }
   XPutImage(did, xfish[4], fgc, &image, 0, 0, 0, 0, rinfo[1].width, rinfo[1].height);

   free(x2);
   finfo = (fish *)malloc(flimit * sizeof(fish));
}


/* ===========================================================================
Toggle secure mode on receipt of signal.
=========================================================================== */
int toggle_secure()
{
   pmode = !pmode;
   if (pmode) XLowerWindow(did, wid); else XRaiseWindow(did, wid);
   XFlush(did);
}


/* ===========================================================================
Shutdown the display server.  Note that while this is not explicitly required,
it is a 'good idea' as it forces a final XSync operation.
=========================================================================== */
void terminate(y)
int y;
{
   XCloseDisplay(did);
   exit(0);
}


/* ===========================================================================
Initialize signal so that SIGUSR1 causes secure mode to toggle.
Also, initialize signal so that SIGTERM causes a proper shutdown.
=========================================================================== */
void init_signals()
{
#ifndef SYSV
   struct sigvec  vec;

   vec.sv_handler = toggle_secure;
   vec.sv_mask    = 0;
   vec.sv_onstack = 0;
   sigvector(SIGUSR1, &vec, &vec);

   vec.sv_handler = terminate;
   sigvector(SIGTERM, &vec, &vec);
#else	
   signal(SIGTERM,terminate);
#endif
}


/* ===========================================================================
Variety of initialization calls, including getting the window up and running.
=========================================================================== */
void initialize()
{
   unsigned long         mask;
   XSetWindowAttributes  atts;

   sid = XDefaultScreen(did);
   width  = XDisplayWidth(did, sid) - 2;
   height = XDisplayHeight(did, sid) - 2;

   init_colormap();
   init_signals();

   mask = CWBackPixel | CWOverrideRedirect;
   atts.background_pixel  = cmap[0];
   atts.override_redirect = True;

   wid = XCreateWindow(did, XRootWindow(did, sid), 1, 1, width, height, 0,
		       CopyFromParent, InputOutput, CopyFromParent, mask, &atts);

   if (blimit) init_bubbles();
   if (flimit) init_fish();
   XStoreName(did, wid, pname);
   XMapWindow(did, wid);
}


/* ===========================================================================
Create a new bubble.  Placement along the x axis is random, as is the size of
the bubble.  Increment value is determined by speed.
=========================================================================== */
void new_bubble(b0)
bubble  *b0;
{
   register int      s;
   register bubble  *b = b0;

   b->x = width * myrand() / RAND_F_MAX;
   for (s = 1; (s == 1) || (s == 3); s = 1.0 + myrand() / RAND_F_1_8);
   b->s = s;
   b->y = height - b->s;
   if ((b->i = smooth * height / (float)binc[s]) == 0) b->i = 1;
   XCopyPlane(did, xbubbles[s], wid, bgc1, 0, 0, s, s, b->x, b->y, 1);
   XCopyPlane(did, xbubbles[s], wid, bgc2, 0, 0, s, s, b->x, b->y, 1);
}


/* ===========================================================================
Clear all bubbles.
=========================================================================== */
void clear_bubbles()
{
   register int      i, s;
   register bubble  *b;

   for (i = 0; i < blimit; i++) {
      b = &binfo[i];
      s = b->s;
      XClearArea(did, wid, b->x, b->y, s, s, False);
   }
}


/* ===========================================================================
Move and draw all bubbles.  Random left-right factor can move bubble one
size-unit in either direction.
=========================================================================== */
void draw_bubbles()
{
   register int      i, j, s;
   register bubble  *b;

   for (i = 0; i < blimit; i++) {
      b = &binfo[i];
      s = b->s;
      if ((b->y -= b->i) > 0) {
	 j = myrand();
         if (j < RAND_I_1_4) b->x -= s; else if (j > RAND_I_3_4) b->x += s;
         XCopyPlane(did, xbubbles[s], wid, bgc1, 0, 0, s, s, b->x, b->y, 1);
         XCopyPlane(did, xbubbles[s], wid, bgc2, 0, 0, s, s, b->x, b->y, 1);
      }
      else {
         new_bubble(b);
      }
   }
}


/* ===========================================================================
Fish collision detection.  The specified fish is checked against all other
fish for overlap.  The xt parameter specifies the x axis tolerance for overlap.
=========================================================================== */
int collide_fish(f0, xt)
fish  *f0;
int    xt;
{
   register int       i, off;
   register fish     *f = f0, *g;
   register rasinfo  *rf = &rinfo[f->s], *rg;

   for (i = 0; i < flimit; i++) {
      if ((g = &finfo[i]) != f) {
         rg = &rinfo[g->s];
         off = g->y - f->y;
         if (((off >= 0) && (off < rf->height)) || ((off < 0) && (-off < rg->height))) {
            off = g->x - f->x;
            if (((off >= 0) && (off < xt * rf->width)) || ((off < 0) && (-off < xt * rg->width))) return(1);
         }
      }
   }
   return(0);
}


/* ===========================================================================
Create a new fish.   Placement along the y axis is random, as is the side
from which the fish appears.  Direction is determined from side.  Increment
is also random.
=========================================================================== */
void new_fish(f0)
fish  *f0;
{
   register int       i, collide;
   register fish     *f = f0;
   register rasinfo  *rf;

   for (i = 0, collide = 1; (i < 16) && (collide); i++) {
      if (myrand() < RAND_I_1_2) f->s = 0; else f->s = 1;
      rf = &rinfo[f->s];
      f->y = (height - rf->height) * myrand() / RAND_F_MAX;
      if ((f->i = smooth * width / (8.0 * (1.0 + myrand() / RAND_F_1_8))) == 0) f->i = 1;
      if (myrand() < RAND_I_1_2) {
         f->d = 1;
         f->x = width;
      }
      else {
         f->d = 2;
         f->x = -rf->width;
      }

      collide = collide_fish(f, 2);
   }

   if (!collide)
      XCopyArea(did, xfish[f->s + f->s + f->d], wid, fgc, 0, 0, rf->width, rf->height, f->x, f->y);
   else
      f->d = 0;
}


/* ===========================================================================
Move all the fish.  Clearing old fish is accomplished by masking only the
exposed areas of the old fish.  Random up-down factor can move fish 1/4 a
fish height in either direction, if no collisions are caused.
=========================================================================== */
void move_fish()
{
   register int       x, y, z;
   int                i, done, ofx, ofy;
   register fish     *f;
   register rasinfo  *rf;

   for (i = 0; i < flimit; i++) {
      f  = &finfo[i];
      rf = &rinfo[f->s];
      if (f->d) {
         ofx = f->x;
         ofy = f->y;

         if (f->d == 1) {
            done = ((f->x -= f->i) < -rf->width);
            x = f->x + rf->width;
         }
         else if (f->d == 2) {
            done = ((f->x += f->i) > width);
            x = f->x - f->i;
         }

	 if (done) {
            XClearArea(did, wid, x, f->y, f->i, rf->height, False);
	    new_fish(f);
	 }
	 else {
	    if (collide_fish(f, 1)) {
	       if ((f->d = 3 - f->d) == 1) {
		  f->x = f->x - 2 * f->i;
		  x = f->x + rf->width; 
	       }
	       else {
		  f->x = f->x + 2 * f->i;
		  x = f->x - f->i;
	       }
	       XCopyArea(did, xfish[f->s + f->s + f->d], wid, fgc, 0, 0, rf->width, rf->height, f->x, f->y);
	       XClearArea(did, wid, x, f->y, f->i, rf->height, False);
	    }
	    else {
	       z = myrand();
               if (z < RAND_I_1_4) y = f->i / 4; else if (z > RAND_I_3_4) y = f->i / -4; else y = 0;
               if (y) {
                  f->y += y;
                  if (collide_fish(f, 1)) {
                     f->y -= y;
                     y = 0;
                  }
                  else {
                     if (y > 0) {
                        z = f->y - y;
		     }
                     else {
		        z = f->y + rf->height;
                        y *= -1;
		     }
                  }
               }

               XCopyArea(did, xfish[f->s + f->s + f->d], wid, fgc, 0, 0, rf->width, rf->height, f->x, f->y);
               XClearArea(did, wid, x, ofy, f->i, rf->height, False);
               if (y) XClearArea(did, wid, ofx, z, rf->width, y, False);
            }
         }
      }
      else {
	 new_fish(f);
      }
   }
}


/* ===========================================================================
Higher-resolution sleep
=========================================================================== */
void high_res_sleep(seconds)
double  seconds;
{
   int             fds = 0;
   struct timeval  timeout;

   timeout.tv_sec  = seconds;
   timeout.tv_usec = (seconds - timeout.tv_sec) * 1000000.0;
   select(0,(fd_set *) &fds,(fd_set *)  &fds,(fd_set *)  &fds, &timeout);
}


/* ===========================================================================
=========================================================================== */
void main(argc, argv)
int     argc;
char  **argv;
{
   int  i;

   pname = argv[0];
   parse(argc, argv);
   initialize();

   srand((unsigned)getpid());
   for (i = 0; i < blimit; i++) new_bubble(&binfo[i]);
   for (i = 0; i < flimit; i++) new_fish(&finfo[i]);
   if (pmode) XLowerWindow(did, wid); else XRaiseWindow(did, wid);
   XFlush(did);

   for (;;) {
      high_res_sleep(delay);
      clear_bubbles();
      move_fish();
      draw_bubbles();
      if (pmode) XLowerWindow(did, wid); else XRaiseWindow(did, wid);
      XFlush(did);
   }
}
